# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2021 Comet ML INC
#  This source code is licensed under the MIT license.
# *******************************************************

""" This module handles asynchronous downloading of the files."""

import logging
import math
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple

from .logging_messages import (
    FILE_DOWNLOAD_MANAGER_COMPLETED,
    FILE_DOWNLOAD_MANAGER_MONITOR_FIRST_MESSAGE,
    FILE_DOWNLOAD_MANAGER_MONITOR_PROGRESSION,
    FILE_DOWNLOAD_MANAGER_MONITOR_PROGRESSION_UNKNOWN_ETA,
)
from .thread_pool import Future, get_thread_pool
from .utils import format_bytes, get_time_monotonic

LOGGER = logging.getLogger(__name__)


class FileDownloadSizeMonitor(object):
    """The monitor callback to maintain the file download progress"""

    __slots__ = ["total_size", "bytes_written"]

    def __init__(self) -> None:
        self.total_size = None
        self.bytes_written = 0

    def monitor_callback(self, bytes_written: int) -> None:
        if bytes_written is not None:  # Python2.7 compatibility check
            self.bytes_written += bytes_written


class DownloadResult(object):
    """The multiprocessing.pool.AsyncResult wrapper providing additional information about download task"""

    def __init__(self, future: Future, monitor: FileDownloadSizeMonitor) -> None:
        self.future = future
        self.monitor = monitor

    def ready(self) -> bool:
        """Allows to check if wrapped Future successfully finished"""
        return self.future.done()

    def get(self, timeout: Optional[int] = None) -> Any:
        """Return the result when it arrives from the wrapped Future.
        If timeout is not None and the result does not arrive within timeout seconds then
        TimeoutError is raised. If the remote call raised an exception then
        that exception will be reraised by get()."""
        return self.future.result(timeout)


class FileDownloadManager(object):
    """The manager to handle downloading of the files and reporting the progress"""

    def __init__(
        self, worker_cpu_ratio: int, worker_count: Optional[int] = None
    ) -> None:
        """Creates new instance with specified ration of workers per CPU. This will instantiate the thread pool for
        asynchronous downloads processing in multiple processes.
        Args:
            worker_cpu_ratio: int - the number of workers per CPU
        """
        self.download_results = []  # type: List[DownloadResult]

        pool_size, cpu_count, self._executor = get_thread_pool(
            worker_cpu_ratio, worker_count
        )

        LOGGER.debug(
            "FileDownloadManager instantiated with %d threads, %d CPUs, %d worker_cpu_ratio, %s worker_count",
            pool_size,
            cpu_count,
            worker_cpu_ratio,
            worker_count,
        )

    def download_file_async(
        self,
        download_func: Callable,
        estimated_size: Optional[int] = None,
        **kwargs: Optional[Any]
    ) -> DownloadResult:
        """Registers file to be downloaded asynchronously using specified download processing function.
        Args:
            download_func: the function to maintain file downloading routines.
            estimated_size: Optional, int - the estimated size of the file to be downloaded (bytes).
        """
        monitor = FileDownloadSizeMonitor()
        if estimated_size is not None:
            monitor.total_size = estimated_size

        kwargs["_monitor"] = monitor
        future = self._executor.submit(download_func, **kwargs)
        result = DownloadResult(future=future, monitor=monitor)
        self.download_results.append(result)
        return result

    def all_done(self) -> bool:
        """Allows to check if all downloads completed"""
        return all(result.ready() for result in self.download_results)

    def remaining_data(self) -> Tuple[int, int, int]:
        """Calculates the number of remaining files, bytes, and total size to
        be downloaded."""
        remaining_downloads = 0
        remaining_bytes_to_download = 0
        total_size = 0
        for result in self.download_results:
            monitor = result.monitor
            if monitor.total_size is None or monitor.bytes_written is None:
                continue

            total_size += monitor.total_size

            if result.ready() is True:
                continue

            remaining_downloads += 1
            remaining_bytes_to_download += monitor.total_size - monitor.bytes_written

        return remaining_downloads, remaining_bytes_to_download, total_size

    def remaining_downloads(self) -> int:
        """Returns number of files to be downloaded"""
        status_list = [result.ready() for result in self.download_results]
        return status_list.count(False)

    def close(self) -> None:
        self._executor.close()

    def join(self) -> None:
        self._executor.join()


class FileDownloadManagerMonitor(object):
    """The monitor to log download progress of the associated FileDownloadManager"""

    def __init__(self, file_download_manager: FileDownloadManager) -> None:
        self.file_download_manager = file_download_manager
        self.last_remaining_bytes = 0
        self.last_remaining_downloads_log_time = None  # type: Optional[float]

    def log_remaining_downloads(self) -> None:
        (
            downloads,
            remaining_bytes,
            total_size,
        ) = self.file_download_manager.remaining_data()

        current_time = get_time_monotonic()

        if remaining_bytes == 0:
            LOGGER.info(FILE_DOWNLOAD_MANAGER_COMPLETED)
        elif self.last_remaining_downloads_log_time is None:
            LOGGER.info(
                FILE_DOWNLOAD_MANAGER_MONITOR_FIRST_MESSAGE,
                downloads,
                format_bytes(remaining_bytes),
                format_bytes(total_size),
            )
        else:
            processed_bytes = self.last_remaining_bytes - remaining_bytes
            time_elapsed = current_time - self.last_remaining_downloads_log_time
            throughput = processed_bytes / time_elapsed

            if processed_bytes <= 0:
                LOGGER.info(
                    FILE_DOWNLOAD_MANAGER_MONITOR_PROGRESSION_UNKNOWN_ETA,
                    downloads,
                    format_bytes(remaining_bytes),
                    format_bytes(total_size),
                    format_bytes(throughput),
                )
            else:
                # Avoid displaying 0s, also math.ceil returns a float in Python 2.7
                remaining_time = str(int(math.ceil(remaining_bytes / throughput)))

                LOGGER.info(
                    FILE_DOWNLOAD_MANAGER_MONITOR_PROGRESSION,
                    downloads,
                    format_bytes(remaining_bytes),
                    format_bytes(total_size),
                    format_bytes(throughput),
                    remaining_time,
                )

        self.last_remaining_bytes = remaining_bytes
        self.last_remaining_downloads_log_time = current_time

    def all_done(self) -> bool:
        """Checks if all downloads completed by monitored download manager."""
        return self.file_download_manager.all_done()


class DownloadResultHolder(NamedTuple):
    download_result: DownloadResult
    asset_filename: str
    asset_path: str
    asset_metadata: Dict[str, Any]
    asset_id: str
    asset_synced: bool
    asset_type: str
    asset_remote_uri: str
